package com.me.common.ecs;

import com.badlogic.gdx.utils.Array;
import com.me.common.ecs.event.Event;
import com.me.common.ecs.event.Listener;

import java.util.HashMap;
import java.util.Map;

public class Engine {

    private Array<Entity> entities;
    private Array<Entity> toActivate;
    private Array<Entity> toDeactivate;
    private Array<Entity> toRemove;
    private Array<Entity> removedEntities;

    private ComponentBag[] components;
    private Array<BaseSystem> systems;

    private ListenerRegistry listenerRegistry;

    private Map<Class<? extends Component>, ComponentMapper> componentMappers;

    public Engine() {
        this.entities = new Array<>();
        this.toActivate = new Array<>();
        this.toDeactivate = new Array<>();
        this.toRemove = new Array<>();
        this.removedEntities = new Array<>(false, 16);

        this.systems = new Array<>();
        this.listenerRegistry = new ListenerRegistry();
        this.componentMappers = new HashMap<>();
    }

    public void registerComponentClass(Class<? extends Component> clazz) {
        ComponentType.registerComponentType(clazz);
    }

    public void registerSystem(EntitySystem system) {
        this.systems.add(system);
    }

    public void registerListener(Listener listener) {
        listenerRegistry.registerListener(listener);
    }

    public void ready() {
        this.components = new ComponentBag[ComponentType.getRegisteredComponentTypeCount()];
        for (int i = 0; i < components.length; i++) {
            components[i] = new ComponentBag();
        }
    }

    public Array<Entity> getEntities() {
        return entities;
    }

    public int getEntityCount() {
        return entities.size;
    }

    public void update(float dt) {
        activatePending();
        deactivatePending();
        removePending();

        for (BaseSystem system : systems) {
            system.preProcess();
            system.process(dt);
            system.postProcess();
        }
    }

    public Entity createEntity() {
        Entity entity;
        if (!removedEntities.isEmpty()) {
            entity = removedEntities.removeIndex(0);
            entity.reset();
        } else {
            entity = new Entity(this);
        }
        entity.updateUniqueId();
        entities.add(entity);
        return entity;
    }

    protected void removeEntity(Entity entity) {
        toRemove.add(entity);
    }

    private void removeEntityFromSystems(Entity entity) {
        for (BaseSystem system : systems) {
            if (!(system instanceof EntitySystem)) {
                continue;
            }

            EntitySystem entitySystem = (EntitySystem) system;
            if (entitySystem.interestedIn(entity)) {
                entitySystem.removeEntity(entity);
            }
        }
    }

    private void addEntityToSystems(Entity entity) {
        for (BaseSystem system : systems) {
            if (!(system instanceof EntitySystem)) {
                continue;
            }

            EntitySystem entitySystem = (EntitySystem) system;
            if (entitySystem.interestedIn(entity)) {
                entitySystem.addEntity(entity);
            }
        }
    }

    private void removePending() {
        if (toRemove.isEmpty()) {
            return;
        }

        for (Entity entity : toRemove) {
            removeAllEntityComponents(entity.id);
            entities.removeValue(entity, true);
            removeEntityFromSystems(entity);
            removedEntities.add(entity);
        }

        toRemove.clear();
    }

    protected void activateEntity(Entity entity) {
        toActivate.add(entity);
    }

    private void activatePending() {
        if (toActivate.isEmpty()) {
            return;
        }

        for (Entity entity : toActivate) {
            addEntityToSystems(entity);
        }
        toActivate.clear();
    }

    protected void deactivateEntity(Entity entity) {
        toDeactivate.add(entity);
    }

    private void deactivatePending() {
        if (toDeactivate.isEmpty()) {
            return;
        }

        for (Entity entity : toDeactivate) {
            removeEntityFromSystems(entity);
        }
        toDeactivate.clear();
    }

    public void callEvent(Event event) {
        listenerRegistry.callEvent(event);
    }

    private void removeAllEntityComponents(int entityId) {
        for (int i = 0; i < components.length; i++) {
            components[i].insert(entityId, null);
        }
    }

    protected void addEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.id].insert(entity.id, component);
        entity.addComponentType(type);
    }


    protected void removeEntityComponent(Entity entity, Component component) {
        ComponentType type = ComponentType.getComponentType(component.getClass());
        components[type.id].remove(entity.id);
        entity.removeComponentType(type);
    }

    protected <T extends Component> T getEntityComponent(Entity entity, Class<T> clazz) {
        return clazz.cast(components[ComponentType.getTypeId(clazz)].get(entity.id));
    }

    protected Component getEntityComponent(Entity entity, ComponentType type) {
        return components[type.id].get(entity.id);
    }

    protected boolean entityHasComponent(Entity entity, ComponentType type) {
        return components[type.id].contains(entity.id);
    }

    @SuppressWarnings("unchecked")
    public <T extends Component> ComponentMapper<T> getComponentMapper(Class<T> typeClass) {
        ComponentMapper<T> mapper = componentMappers.get(typeClass);
        if (mapper == null) {
            mapper = new ComponentMapper<>(this, typeClass);
            componentMappers.put(typeClass, mapper);
        }
        return mapper;
    }

}
